; GIF LITE v3.0 by Willow
; Written in pure assembler by Ivushkin Andrey aka Willow
; Modified by Diamond
; Modified for MeView by Nable
; This include file contains functions to handle GIF image format
;
; Created: August 15, 2004
; Last changed: January 26, 2009
struc GIF_list
{
    NextImg rd 1
    Left   rw 1
    Top    rw 1
    Width  rw 1
    Height rw 1
    Delay  rd 1
    Displacement rd 1  ; 0 = not specified
		       ; 1 = do not dispose
		       ; 2 = restore to background color
		       ; 3 = restore to previous
    Image rd 1
}

struc GIF_info
{
    Left   rw 1
    Top    rw 1
    Width  rw 1
    Height rw 1
    Delay  rd 1
    Displacement rd 1
    Palette rd 1
}

;*************************************************
;  FUNCTION GetGIFinfo - retrieve Nth image info
;*************************************************
GetGIFinfo:
    mov   ecx,[frame_num]
    mov   esi,[___cur_info]
    mov   ebx,esi
    mov   edi,IMG_INFO_Block
    jecxz ____eloop
@@:
    mov   esi,[esi]
    add   esi,ebx
    loop  @b
____eloop:
    lodsd
    movsd
    movsd
    movsd
    movsd
    lodsd
    add  eax,ebx
    mov  [soi],eax
    mov  [___Palette],esi
    push eax
    sub  [esp],esi
    fld1
    fild dword[esp]
    fyl2x
    fistp dword[esp]
    pop  eax
    dec  eax
    dec  eax
    mov  [IMG_INFO_BitsPerPointReal],ax
    ret

_null fix 0x1000

;****************************************
;  FUNCTION ReadGIF - unpacks GIF image
;****************************************
;in:
;   eax - pointer to file
;   [___cur_info] - pointer to output area
ReadGIF:
    push dword[___cur_info]

    mov  esi,[esp]
    neg  esi
    mov  [output_data_size],esi ;output_data_size = -start_of_data
				;when we reach the end we'll add
				;end_of_data value to get the size
    mov  esi,eax
    xor  eax,eax
    inc  eax
    mov  ecx,[esi+0xa]
    add  esi,0xd
    mov  edi,esi
    test cl,cl
    jns  ReadGIF___nextblock
    mov  [___globalColor],esi
    call ReadGIF___Gif_skipmap
ReadGIF___nextblock:
    cmp  byte[edi],0x21
    jne  ReadGIF___noextblock
    inc  edi
    cmp  byte[edi],0xf9 ; Graphic Control Ext
    jne  ReadGIF___no_gc
    movzx eax,word [edi+3]
    mov  [___anim_delay],eax
    mov  al,[edi+2]
    shr  al,2
    and  eax,7
    mov  [___anim_disp],eax
    add  edi,7
    jmp  ReadGIF___nextblock
ReadGIF___no_gc:
    inc  edi
ReadGIF___block_skip:
    movzx eax,byte[edi]
    lea  edi,[edi+eax+1]
    test eax,eax
    jnz  ReadGIF___block_skip
    jmp  ReadGIF___nextblock
ReadGIF___noextblock:
    mov  al,8
    cmp  byte[edi],0x2c    ; image beginning
    je	 ReadGIF___supported_block
    pop  dword[___cur_info]
    jmp  unsupported
ReadGIF___supported_block:
    inc  [___img_count]
    inc  edi
    mov  esi,[___cur_info]
    xchg esi,edi
    scasd
    movzx ecx,word[esi+4]
    mov  [___width],ecx
    movzx eax,word[esi+6]
    imul eax,ecx
    mov  [___img_end],eax
    inc  eax
    mov  [___row_end],eax
    and  [___pass],0
    test byte[esi+8],40h
    jz	 @f
    mov  [___row_end],ecx
@@:
    movsd
;-------------------------------------------------------------------------------
;Nable's add: memory resizing
;Now at [esi] is a dword with this format: Width+Height*65536, i.e. sizes of
;the image that we'll unpack. I'll use this data to allocate memory
    lodsd
    stosd
    movzx ecx,ax
    shr  eax,16
    mul  ecx
    add  eax,4*256+8*4 ;maximal possible space for palette and descriptor.
;Unused space won't be wasted.It'll be used for next image or freed in cleanup
    add  eax,edi       ;eax = new memory size req.
    sub  eax,[_program_memory] ;calculate how much to add. We add, not resize
    jc	 ReadGIF___have_enough_space ;Maybe we have already enough?
    push eax	       ;Add if not
    call mem_add_pages

ReadGIF___have_enough_space:
;-------------------------------------------------------------------------------
;    movsd
    mov  eax,[___anim_delay]
    stosd
    mov  eax,[___anim_disp]
    stosd
    push edi
    mov  ecx,[esi]
    inc  esi
    test cl,cl
    js	 ReadGIF___uselocal
    push [___globalColor]
    mov  edi,esi
    jmp  ReadGIF___setPal
ReadGIF___uselocal:
    call ReadGIF___Gif_skipmap
    push esi
ReadGIF___setPal:
    movzx ecx,byte[edi]
    inc  ecx
    mov  [___codesize],ecx
    dec  ecx
    lea  esi,[edi+1]
    mov  edi,hashtable
    xor  eax,eax
    lodsb		; eax - block_count
    add  eax,esi
    mov  [___block_ofs],eax
    mov  [___bit_count],8
    mov  eax,1
    shl  eax,cl
    mov  [___CC],eax
    mov  ecx,eax
    inc  eax
    mov  [___EOI],eax
    mov  eax, _null shl 16
ReadGIF___filltable:
    stosd
    inc  eax
    loop ReadGIF___filltable
    pop  eax
    pop  edi
    push edi
    scasd
    push esi
    mov  esi,eax
    mov  ecx,[___CC]
@@:
    lodsd
    dec  esi
    bswap eax
    shr  eax,8
    stosd
    loop @b
    pop  esi
    pop  eax
    mov  [eax],edi
    mov  [___img_start],edi
    add  [___img_end],edi
    add  [___row_end],edi
ReadGIF___reinit:
    mov  edx,[___EOI]
    inc  edx
    push [___codesize]
    pop  [___compsize]
    call ReadGIF___Gif_get_sym
    cmp  eax,[___CC]
    je	 ReadGIF___reinit
    call ReadGIF___Gif_output
ReadGIF___cycle:
    movzx ebx,ax
    call ReadGIF___Gif_get_sym
    cmp  eax,edx
    jae  ReadGIF___notintable
    cmp  eax,[___CC]
    je	 ReadGIF___reinit
    cmp  eax,[___EOI]
    je	 ReadGIF___end
    call ReadGIF___Gif_output
ReadGIF___add:
    mov  dword [hashtable+edx*4],ebx
    cmp  edx,0xFFF
    jae  ReadGIF___cycle
    inc  edx
    bsr  ebx,edx
    cmp  ebx,[___compsize]
    jne  ReadGIF___noinc
    inc  [___compsize]
ReadGIF___noinc:
    jmp  ReadGIF___cycle
ReadGIF___notintable:
    push eax
    mov  eax,ebx
    call ReadGIF___Gif_output
    push ebx
    movzx eax,bx
    call ReadGIF___Gif_output
    pop  ebx eax
    jmp  ReadGIF___add
ReadGIF___end:
    mov  edi,[___img_end]
    mov  eax,[___cur_info]
    mov  [eax],edi
    mov  [___cur_info],edi
    inc  esi
    inc  esi
    xchg esi,edi
ReadGIF___nxt:
    cmp  byte[edi],0
    jnz  ReadGIF___continue
    inc  edi
    jmp  ReadGIF___nxt
ReadGIF___continue:
    cmp  byte[edi],0x3b
    jne  ReadGIF___nextblock
    and  dword [eax],0		;this is the end of everything
    add  [output_data_size],esi ;output_data_size = - start_of_data + end_of_data 
    mov  byte[IMG_INFO_BitsPerPoint],8
    pop  dword[___cur_info]
    call make_multi_image_structure_relocatable
;now I'll shrink original file's data - no need in it, then dealloc. its mem.
    mov  ecx,[output_data_size]
;    shr  ecx,2              ;see comment near 'rep movsb'
;    inc  ecx                ;bytes -> dwords
    mov  edi,I_END
    mov  esi,___cur_info
    push dword[esi]	    ;<-mov esi,[___cur_info] (old value) (stage1)
    mov  [esi],edi		;mov [___cur_info],I_END (new value)
    pop  esi		    ;<-mov esi,[___cur_info] (old value) (stage2)
    rep  movsb		    ;I can use shr ecx,2 ,rep movsd for time, but
			    ;I think it's not needed.
    push edi
    call mem_resize_everything ;now free unused memory
    call GetGIFinfo
    jmp  converted

;-------------------------------------------------------------------------------
ReadGIF___Gif_skipmap:
; in: ecx - image descriptor, esi - pointer to colormap
; out: edi - pointer to area after colormap

    and  ecx,111b
    inc  ecx		; color map size
    m2m  ebx,1
    shl  ebx,cl
    lea  ebx,[ebx*2+ebx]
    lea  edi,[esi+ebx]
    ret

ReadGIF___Gif_get_sym:
    mov  ecx,[___compsize]
    push ecx
    xor  eax,eax
___Gif_get_sym___shift:
    ror  byte[esi],1
    rcr  eax,1
    dec  [___bit_count]
    jnz  ___Gif_get_sym___loop1
    inc  esi
    cmp  esi,[___block_ofs]
    jb	 ___Gif_get_sym___noblock
    push eax
    xor  eax,eax
    lodsb
    test eax,eax
    jnz  ___Gif_get_sym___nextbl
    mov  eax,[___EOI]
    dec  esi
    dec  esi
    add  esp,8
    jmp  ___Gif_get_sym___exx
___Gif_get_sym___nextbl:
    add  eax,esi
    mov  [___block_ofs],eax
    pop  eax
___Gif_get_sym___noblock:
    mov  [___bit_count],8
___Gif_get_sym___loop1:
    loop ___Gif_get_sym___shift
    pop  ecx
    rol  eax,cl
___Gif_get_sym___exx:
    xor  ecx,ecx
    ret

ReadGIF___Gif_output:
    push esi eax edx
    mov  edx,hashtable
___Gif_output___next:
    push word[edx+eax*4]
    mov  ax,word[edx+eax*4+2]
    inc  ecx
    cmp  ax,_null
    jnz  ___Gif_output___next
    shl  ebx,16
    mov  bx,[esp]
___Gif_output___loop2:
    pop  ax

    stosb

    cmp  edi,[___row_end]
    jb	 ___Gif_output___norowend
    mov  eax,[___width]
    push eax
    sub  edi,eax
    add  eax,eax
    cmp  [___pass],3
    jz	 @f
    add  eax,eax
    cmp  [___pass],2
    jz	 @f
    add  eax,eax
@@:
    add  edi,eax
    pop  eax
    cmp  edi,[___img_end]
    jb	 ___Gif_output___nextrow
    mov  edi,[___img_start]
    inc  [___pass]
    add  edi,eax
    cmp  [___pass],3
    jz	 @f
    add  edi,eax
    cmp  [___pass],2
    jz	 @f
    add  edi,eax
    add  edi,eax
@@:
___Gif_output___nextrow:
    add  eax,edi
    mov  [___row_end],eax
    xor  eax,eax
___Gif_output___norowend:

    loop ___Gif_output___loop2
    pop  edx eax esi
    ret