;-------------------------------------------------------------------------------
;Targa (TGA) image files decoder by Nable. A lot of ideas were taken from
;Hiview 1.2 by Mohammad A. REZAEI
;
;While MeView determines the type of the file and if if it's a TGA file, then
;the following registers will be set:
;cl=cmatype (colormap presented or not)
;ch=subtype
;dl=bpp
;dh=flags
;eax=pointer to file
;ebx=size of file
;During the decoding ebp+(one ram page) is the limit of your ram!Save the ram!
;Last changed: 26.01.2009
;-------------------------------------------------------------------------------
decode_tga:
    push dword[eax+12]
    movzx ebp,word[esp]  ;ebp=width
    pop  dword[IMG_INFO_Sizes]
    mov  byte[IMG_INFO_BitsPerPoint],dl
    movzx esi,byte[eax]
    lea  esi,[eax+esi+18]
    test cl,cl		 ;cl==1 if colormap presented else cl==0
    jnp  tga_skip_colormap
    cmp  ch,8
    jb	 tga_colormap_skipped
;-------------------------------------------------------------------------------
    lea  edi,[eax+ebx]
    mov  [soi],edi
    mov  [start_of_img_data],esi
;-------------------------------------------------------------------------------
    jmp  tga_colormap_skipped_rle
tga_skip_colormap:
    mov  [___Palette],esi
    push ecx
    mov  cl,byte[eax+7]  ;cl=size of colormap entries in bits
@@:
    add  si,[eax+5]	 ;number of entries in colormap
    sub  cl,8
    jnz  @b		 ;esi points to the beginning of the image data

    mov  edi,esi
    sub  edi,[___Palette];edi = sizeof(old_palette)

    lea  ecx,[eax+ebx]	 ;ecx points after the file
    push ecx
    sub  ecx,esi	 ;ecx = sizeof(packed_image_data)
    pop  esi		 ;esi points after the file

    push edi		 ;push sizeof(old_palette)
    movzx edi,word[eax+5];number of entries in colormap
    push edi
    shl  edi,2		 ;edi now contains a size of new palette
    sub  edi,[esp]	 ;edi = sizeof(new_palette) - sizeof(old_palette)
    add  edi,esi	 ;edi = end_of_image_data+additional_space_for_palette
    pop  dword[esp]	 ;add esp,4 , [esp]=[esp-4]

;------
    cmp  byte[esp+4+1],8   ;ecx is in stack
    jb	 @f		 ;if it is rle...
    mov  [soi],edi ;...unpacked image data will be at (edi) (at some first time)
@@:
;------

    inc  ecx
    std
    rep  movsb
    cld

    pop  ecx		 ;ecx=number of entries in colormap
    call make_palette_32
    pop  ecx
    inc  edi
    mov  esi,edi
    mov  [start_of_img_data],edi
tga_colormap_skipped:
    mov  edi,esi	 ;if no colormap it is needed
    cmp  ch,8
    ja	 tga_colormap_skipped_rle
    mov  [soi],esi
tga_colormap_skipped_rle:
    shr  dl,3		 ;dl=pixel_size
    ror  edx,16
    mov  dx,cx
    ror  edx,16
    push edx
    movzx ecx,word[eax+14];height of image
    imul ebp,ecx	;now ebp contains the size of unpacked image in pixels
    mov  [___raw_size],ebp
    cmp  dl,2		 ;if it is 16(15)bpp we have to make it 24bpp
    jne  @f
    inc  edx
@@:
    movzx edx,dl	 ;edx=output_pixel_size
    imul ebp,edx
    lea  ebp,[ebp+esi+2048];end_of_image=sizeof(unpacked_image)+soi+reserve
;ebp is a limit that alarms the decoder when reached and then we
;free unused ram and use it
;------
    push ebp
    call mem_resize_everything
;------

    mov  edx,[esp]
    pop  dword[tga_file_information1]

    mov  edi,[soi]

tga_first_step: 	 ;unpack the image if it needed
    cmp  byte[tga_subtype],8 ;rle packed image?
    jb	 tga_second_step ;if not this step is ended
    sub  byte[tga_subtype],8
@@:
    call tga_read_rle_pixel
    call tga_put_pixel
    dec  dword[___raw_size]
    jnz  @b

;------
    call save_ram
    push dword[start_of_img_data]
    pop  dword[soi]
;------

tga_second_step:	 ;de-interlace the image if it needed
    cmp  byte[tga_subtype],3 ;gray scale image?
    jne  tga_not_a_grayscale_image
    mov  ecx,[IMG_INFO_Sizes]
    movzx eax,cx
    shr  ecx,16
    imul ecx,eax	 ;ecx=size of image in pixels
    mov  esi,soi
    push dword[esi]
    movzx eax,word[I_END+5] ;number of entries in palette
    push eax
    shl  eax,2		 ;size of palette bytes
    add  dword[esi],eax
    pop  eax		 ;eax=number of entries in palette
    mov  edi,[esi]
    mov  esi,[esp]
    pop  dword[___Palette]
    push eax
    add  edi,ecx
    add  esi,ecx
    dec  edi
    dec  esi
    std
    rep  movsb
    cld
    pop  ecx
    mov  edi,[___Palette]
    xor  eax,eax	 ;begin from black
@@:
    stosd
    add  eax,0x010101
    loop @b		 ;build the palette

tga_not_a_grayscale_image:
;test_if_deinterlacing_needed:
    ;de-interlacing is not supported yet
    jmp  converted

align 4
tga_read_rle_pixel:
    cmp  dword[dup_pixel_count],0 ;Duplicate previously read pixel?
    ja	 duplicate_previously_read_pixel
    mov  ebx,tga_block_count
    dec  dword[ebx]		  ;Decrement pixels remaining in block
    jns  read_non_rle_pixel
    xor  eax,eax
    lodsb
    test al,al			  ;Start of duplicate-pixel block?
    jns  .2
    and  al,0x7f
    mov  dword[dup_pixel_count],eax ;Number of duplications after this one
    and  dword[ebx],0		  ;Then read new block header
    jmp  read_non_rle_pixel
.2:
    mov  dword[ebx],eax
read_non_rle_pixel:
    push edx
    mov  edx,[tga_file_information1]
    xor  eax,eax
    mov  dh,dl
@@:
    shl  eax,8
    lodsb
    dec  dl
    jnz  @b
    cmp  dh,3
    jne  @f
    bswap eax
    shr  eax,8
@@:
    mov  [current_pixel],eax
    pop  edx
    ret
duplicate_previously_read_pixel:
    dec  dword[dup_pixel_count]
    ret
align 4
tga_put_pixel:
    mov  eax,[current_pixel]
    mov  edx,[tga_file_information1]
    cmp  edi,ebp
    jb	 @f
    call save_ram
@@:
    cmp  dl,2		 ;if it is 16(15)bpp we have to make it 24bpp
    jne  @f
    push ebx
    mov  ebx,table5to8bits
    shl  eax,3
    shr  al,3
    xlatb
    stosb
    shr  eax,5
    shr  al,3
    xlatb
    stosb
    shr  eax,5
    shr  al,3
    xlatb
    stosb
    pop  ebx
    ret
@@:
    stosb
    shr  eax,8
    dec  dl
    jnz  @b
    ret