;-------------------------------------------------------------------------------
;Z80 screen-shots decoder by Nable. Thanks a lot Aleksandr Bogomaz aka Albom
;for provided documentation.
;Started:      August 7, 2008
;Last changed: January 26, 2009
;-------------------------------------------------------------------------------
;Entry (MeView specific):
;eax - pointer to file
;ebx - file size
;edi = (eax+ebx+4095)&0xFFFFF000 - possible output area
;-------------------------------------------------------------------------------
;  ࠭  ZX Spectrum ⠢ 256 祪  ਧ⠫  192 
;⨪. 樮 ⥬ 뢮 ᨬ  ࠭ ⢫ 
;⠭  8x8 祪.  ࠧ, ࠭   24
;ᨬ ப  32   .
;
;  뢮 ࠦ ᯮ ࠭   ꥬ 6 
; ਡ⮢ 768 .   ࠭  । ﭨ
;8 ᯮ ਧ⠫쭮   㣮 祪 ( 7 ᮮ⢥ 
;窥,  0 - ࠢ). ᫨ -    ⠭,  窠,
;।塞 ⨬ ⮬, ਭ 梥 ⮭, ᫨  襭 -  梥 䮭.
;, ᮮ⢥騥 ப ࠭, ᯮ    ᫥⥫쭮.
; 32  ࠭   ᫥ ࠢ  ப,
;᫥騥 32  - 9- ᢥ, ᫥騥 - 17-   १ 8 ப 
; 57-. ⥬ ᯮ , ᮮ⢥騥 ப 2-, 10-  . . 
;58-.  ࠧ,  ᬥ饭    蠣  8 ப  孨 64
;ப,    ࠭. 筮, 稭 ᮮ⢥⢥  65-  129-
;ப, । ।    ࠭.
;  ZX Spectrum ਡ (梥 ⮭, 梥 䮭, મ  栭)  
; ⤥쭮 窨,   .     ਡ⮢
;ᮮ⢥  ,  ஬:
;
; DO...D2 । 梥 ⮭;
; D3...D5 । 梥 䮭;
; D6 ⠭  ० 襭 મ ( ६  
;⮭,   䮭);
; D7 ⠭  ० 栭.

; ਡ⮢ ᯮ   ᫥⥫쭮.  32 
;ਡ⮢ ᮮ⢥ 孥 ᨬ쭮 ப, ᫥騥 32 - ன
;ᨬ쭮 ப ᢥ     ࠭.
;-------------------------------------------------------------------------------
;During the decoding:
;bl - PixelLeft (this means how much pixels left to put in current string)
;bh - CurrentString
;High half of ebx - use DualStos (two frames per one pixel_write)
;cl - PixelColorIndexInPalette
;ch - BackgroundColorIndexInPalette
;High half of ecx - blinking flag
;edx - address of current attribute byte
;-------------------------------------------------------------------------------
_z80_PixColor	equ 000111b
_z80_BgrColor	equ 111000b
_z80_BrightFlag equ (1 shl 6)
_z80_BlinkFlag	equ (1 shl 7)
_z80_AttrString equ 32
_z80_OutputSize equ 256*192+16*4 ;this is minimal output size, if we have
;blinking pixels, then output size will be _z80_FullFrame*2
_z80_FullFrame	equ (_z80_OutputSize+6*4) ;6*4 - descriptor size

;-------------------------------------------------------------------------------
decode_z80:
;Initialisation
    mov  byte[IMG_INFO_BitsPerPoint],8
    mov  dword[IMG_INFO_Sizes],192*65536+256 ;height shl 16 + width
    m2m  [___img_count],1 ;if there are no blinking pixels then we have only a
			  ;single frame, else we have two frames.
    push (_z80_FullFrame*2)
    call mem_add_pages	  ;we allocate the maximal value that can be needed,
    ;unused memory will be returned to a free pool after unpacking

    mov  esi,eax	  ;esi - source pointer, edi - destination
    lea  edi,[eax+ebx]	  ;don't care about alignment
    mov  [___cur_info],edi;[___cur_info] - pointer to output area

;-------------------------------------------------------------------------------
;At first we'll skip 17byte comment that's attached to 6929byte screens. First
;11 bytes of this comment - plain text. I'm planing to add to MeView support of
;attachments - so...
    cmp  ebx,6929
    jne  @f
    add  esi,17 ;...but now we just skip it.
@@:
    push esi		  ;it'll be poped to edi in decode_z80_finish
;-------------------------------------------------------------------------------
;2nd stage: we determine if there are any blinking pixels (if no - we'll produce
;statical single image)
    mov  ecx,768
    lea  edx,[eax+ebx-768];edx points to attribute area
    xor  ebx,ebx	  ;begin from <0,0> (for further decoding)
@@:
    test byte[edx+ecx-1],_z80_BlinkFlag ;such addressing is a good optimisation
					;(as I hope), edx is unchanged
    jnz  decode_z80_with_blinking
    loop @b
decode_z80_without_blinking:
    mov  [output_data_size],_z80_OutputSize
    mov  [___Palette],esi
    push esi
    add  esi,16*4
    mov  [soi],esi
    pop  esi
    call write_16color_palette		;see const.inc
    jmp  decode_z80_main_stage

decode_z80_with_blinking:
    inc  dword[___img_count]
    mov  [output_data_size],_z80_FullFrame*2
    or	 ebx,0xFFFF0000
;set initial pointers
    push esi
    add  esi,6*4
    mov  [___Palette],esi
    add  esi,16*4
    mov  [soi],esi
    pop  esi
;build descriptors
    mov  eax,_z80_FullFrame
    and  dword[edi+eax],0		;setup NextImage pointers (frame #2)
    stosd				;setup NextImage pointers (frame #1)
    and  dword[edi],0			;left and top offsets = 0 (frame #1)
    and  dword[edi+eax],0		;left and top offsets = 0 (frame #2)
    scasd				;add edi,4
    mov  dword[edi],192*65536+256	;set width and height (frame #1)
    mov  dword[edi+eax],192*65536+256	;set width and height (frame #2)
    scasd				;add edi,4
    m2m  dword[edi],50			;set animation delay to 0.5s (frame #1)
    m2m  dword[edi+eax],50		;set animation delay to 0.5s (frame #2)
    scasd				;add edi,4
    and  dword[edi],0			;displacement = 0 (frame #1)
    and  dword[edi+eax],0		;displacement = 0 (frame #2)
    scasd				;add edi,4
    m2m  dword[edi],16*4+6*4		;palette pointer (frame #1)
    mov  dword[edi+eax],16*4+6*4+_z80_FullFrame ;palette pointer (frame #2)
    scasd				;add edi,4
    call write_16color_palette		;write palette (frame #1)
    lea  edi,[edi+eax-16*4]
    call write_16color_palette		;write palette (frame #2)
    sub  edi,eax
;-------------------------------------------------------------------------------
decode_z80_main_stage:
;3rd stage - convert z80 screen to 8bpp image with palette
decode_z80_main_stage_main_loop:
    test bl,7
    jnz  decode_z80_main_stage_put_now

_z80_update_attributes:
    movsx ecx,byte[edx]   ;note that BlinkFlag is the highest bit in attribute
			  ;byte, so ecx's highest bit is set automatically
    shl  ecx,5
    shr  cl,5
    and  ch,0xF
    test ch,1000b
    jz	 @f
    or	 ecx,1000b	  ;it has the same size with 'or cl,1000b' but could be faster
@@:
    inc  edx

    lodsb
    mov  ah,al

decode_z80_main_stage_put_now:
    shl  ah,1
;-------------------------------------------------------------------------------
_z80_put_pixel:
;In: CF - put pixel with color CL, !CF - pixel with color CH
;High parts of ebx and ecx - as described above
    mov  al,cl		  ;'mov' doesn't affect flags
    jc	 @f
    mov  al,ch
@@:
    stosb		  ;'stosb' doesn't affect flags

    test ebx,ebx
    jns  @f
    test ecx,ecx
    jns  .1
    mov  al,ch
.1:
    mov  [edi+_z80_FullFrame-1],al
@@:
    inc  bl		  ;next pixel
    jz	 decode_z80_main_stage_row_finished
    jmp  decode_z80_main_stage_main_loop
;-------------------------------------------------------------------------------
decode_z80_main_stage_row_finished:
    cmp  bh,191 	  ;is image finished?
    jb	 decode_z80_main_stage_image_not_finished ;no.
decode_z80_finish:
;shrink original file, but left extension if it was
    mov  ecx,[output_data_size]
    pop  edi		  ;do you remember 'push esi'?
    mov  esi,edi
    xchg esi,[___cur_info]
    rep  movsb
;return freed memory to OS
    push edi
    call mem_resize_everything
    jmp  converted	  ;now really finished
;-------------------------------------------------------------------------------
;or not finished yet. Branch according to a row number (see documentation)
decode_z80_main_stage_next_third:
    sub  bh,7
    sub  edi,256*(8-1)
    jmp  decode_z80_main_stage_main_loop

decode_z80_main_stage_image_not_finished:
;next row
    add  bh,8		  ;refer to documentation
    add  edi,256*(8-1)

;if finished row is 63 or 127 then we process next third of the image
    cmp  bh,63+8
    je	 decode_z80_main_stage_next_third
    cmp  bh,127+8
    je	 decode_z80_main_stage_next_third

    cmp  bh,56+8	  ;if finished row in [56;63) or [120;127) or [184;191)
    jb	 decode_z80_main_stage_main_loop
    cmp  bh,63+8
    jb	 .4
    cmp  bh,120+8
    jb	 decode_z80_main_stage_main_loop
    cmp  bh,127+8
    jb	 .4
    cmp  bh,184+8
    jb	 decode_z80_main_stage_main_loop
;note that if we are here then bh is < 191 (see label .2) but >= 184+8
.4:
;and if we here then bh is in [56;63) or [120;127) or [184;191)
    sub  bh,(8+56-1)
    sub  edi,256*(8+56-1)
    sub  edx,_z80_AttrString*8
    jmp  decode_z80_main_stage_main_loop
