;-------------------------------------------------------------------------------
;MeView GetPixel routines. They are used when we want to get one pixel, that is
;at (ECX,EDX) in the image (if we use Prepare2Get then GetPixel) or from [ESI]
;(if we just use GetPixel), in case of images with bpp < 8 from sample number is 
;set by [ESP], i.e. you send it as a stdcall parameter). In case of bpp <= 8,
;ebp points to palette
;In most cases we shouldn't call these routines directly - we invoke them, i.e.
;call Prepare2Get ;Prepare2Get sets register ESI and, may be, pushs sample 
;call [GetPixel]	;number for GetPixel, GetPixel adjusts ESI in case of it was
;last sample in a byte, so after GetPixel ESI points to location of next pixel.
;Started: 28.12.2008
;Last changed: 29.12.2008
;Most part of this file is from kernel sources
;-------------------------------------------------------------------------------
;Prepare2Get routines:
; align 4
; Prepare2Get.1bpp:
	; add	eax, ecx
	; push	ecx
	; add	eax, 7
	; add	ecx, 7
	; shr	eax, 3
	; shr	ecx, 3
	; sub	eax, ecx
	; pop	ecx
	; ret

; align 4
; Prepare2Get.2bpp:
	; add     eax, ecx
	; push    ecx
	; add     ecx, 3
	; add     eax, 3
	; shr     ecx, 2
	; shr     eax, 2
	; sub     eax, ecx
	; pop     ecx
	; ret

; align 4
; Prepare2Get.4bpp:
	; add	eax, ecx
	; push	ecx
	; add	ecx, 1
	; add	eax, 1
	; shr	ecx, 1
	; shr	eax, 1
	; sub	eax, ecx
	; pop	ecx
	; ret

align 4
Prepare2Get:
	movzx	esi,word[IMG_INFO_Width]
	imul	esi,edx
	add		esi,ecx
	imul	esi,[Prepare2Get.Ratio]
	add		esi,[soi]
	ret
;-------------------------------------------------------------------------------
;GetPixel routines:
; align 4
; GetPixel.1bpp:
	; push	edx
;;	mov	edx, [esp+8]
	; mov	al, [edx]
	; add	al, al
	; jnz	@f
	; lodsb
	; adc	al, al
; @@:
	; mov	[edx], al
	; sbb	eax, eax
	; and	eax, [edx+8]
	; add	eax, [edx+4]
	; pop	edx
	; ret

; align 4
; GetPixel.2bpp:
	; push    edx
;;	mov     edx, [esp+8]
	; mov     al, [edx]
	; mov     ah, al
	; shr     al, 6
	; shl     ah, 2
	; jnz     .nonewbyte
	; lodsb
	; mov     ah, al
	; shr     al, 6
	; shl     ah, 2
	; add     ah, 1
; .nonewbyte:
	; mov     [edx], ah
	; movzx   eax, al
    ; mov     eax, [ebp+eax*4]
    ; pop     edx
	; ret

; align 4
; GetPixel.4bpp:
	; push	edx
;;	mov	edx, [esp+8]
	; add	byte [edx], 80h
	; jc	@f
	; movzx	eax, byte [edx+1]
	; mov	edx, [edx+4]
	; and	eax, 0x0F
	; mov	eax, [ebp+eax*4]
	; pop	edx
	; ret
; @@:
	; movzx	eax, byte [esi]
	; add	esi, 1
	; mov	[edx+1], al
	; shr	eax, 4
	; mov	eax, [ebp+eax*4]
	; pop	edx
	; ret

align 4
GetPixel.8bpp:
	movzx	eax, byte [esi]
	mov	eax, [ebp+eax*4]
	inc	esi
	ret

align 4
GetPixel.15bpp:
; 0RRRRRGGGGGBBBBB -> 00000000RRRRR000GGGGG000BBBBB000
	push    ecx edx
	movzx   eax, word [esi]
	add     esi, 2
	mov     ecx, eax
	mov     edx, eax
	and     eax, 0x1F
	and     ecx, 0x1F shl 5
	and     edx, 0x1F shl 10
	shl     eax, 3
	shl     ecx, 6
	shl     edx, 9
	or      eax, ecx
	or      eax, edx
	pop     edx ecx
	ret


align 4
GetPixel.16bpp:
; RRRRRGGGGGGBBBBB -> 00000000RRRRR000GGGGGG00BBBBB000
	push    ecx edx
	movzx   eax, word [esi]
	add     esi, 2
	mov     ecx, eax
	mov     edx, eax
	and     eax, 0x1F
	and     ecx, 0x3F shl 5
	and     edx, 0x1F shl 11
	shl     eax, 3
	shl     ecx, 5
	shl     edx, 8
	or      eax, ecx
	or      eax, edx
	pop     edx ecx
	ret

align 4
GetPixel.24bpp:
	mov	eax, [esi]
	add	esi, 3
	and eax,0xFFFFFF
	ret

align 4
GetPixel.32bpp:
	lodsd
	ret
	