;-------------------------------------------------------------------------------
;MeView IPC routines. Fully by Nable.
;Started: 27.06.2008
;Last changed: 26.01.2009
;-------------------------------------------------------------------------------
;Function 'IPC_name'.
;
;eax=temp_area (there parameters stored)
;parameter line is like that (without brackets):
;'IPCNxxxxxxxxPATH' or 'TNyyxxxxxxxxPATH', where 'xxxxxxxx' - PID of receiver
;in hexadecimal ASCII format (but without '0x' or 'h', 8 characters)
;PATH - path to file, yy - maximal size of needed thumbnail minus 1 (also in 
;hex, so you can get thumbnail with size up to 256x256 (in case of TNFF)).
;IPCN is for full image, TNyy - for thumbnail.
;Examples: IPCN0000007B/hd1/1/somedir/anotherdir/somefile.gif
;		   TN5F0000007B/hd1/1/somedir/anotherdir/somefile.gif

;first message from MV to receiver has the following format:

;//begin of structure
;dd ImageCount  ;+0  ;image count, 1 - one image, 2 - two images, ...
;dd SomeFlags   ;+4  ;reserved for flags
;dd MsgIISize   ;+8  ;size of the second message (with unpacked image data)

;Note #1:
;if ImageCount > 1 then the following 5 fields will contain
;information about the _first_ image.

;dw ImageLeft   ;+12 ;x coord. of left-top image corner.You can use it or not
;dw ImageTop    ;+14 ;y coord. of left-top image corner.You can use it or not
;dw ImageWidth  ;+16 ;no comments
;dw ImageHeight ;+18 ;no comments
;dw ImageBpp    ;+20 ;bits per pixel. See 65th function info

;dw ImageBppReal ;+22 ;real bpp of image. It's always =< ImageBpp. Use for info
;//end of structure

;Note #2:
;so, the size of the _first message_ is 24 bytes , but reserve at least 64
;bytes for IPC buffer if you want to receive it. Unused space you can use
;later for the second message (I separated data in 2 messages to give you a
;chance to work with memory wisely, i.e. when you receive first message (it's
;rather small, don't you find?) you can exactly calculate how much memory
;you need to allocate for IPC buffer to receive the second message)

;Note #3:
;if (ImageCount =< 1) && (ImageBpp =< 8) then _second message_ has the following
;format (C-like style):

;//begin of structure
;DWORD sizeof(Palette[]);
;COLOR Palette[]; //see 65th function info, sizeof(COLOR)==4
;BYTE  image[];
;//end of structure

;if (image_count =< 1)&&(image_bpp >= 15) then _second
;message_ has the following format (C-like style):
;//begin structure
;COLOR image[]; //sizeof(COLOR)==(image_bpp>>3)
;//end structure

;if (image_count > 1) then second message is as a data output like the gif_lite.inc
;produces (refer to the original file with version 3.0) with these 2 changes:
;#1: all pointers are zero-related (this means that data will be as if output area
;for gif_lite was at 0), so if you follow some pointer, don't forget to add address
;of your "second message's rest in peace area" to get right address. for example, if
;pointer says that next image descriptor (don't mess with GIF's image descriptors)
;is at 0x230 and you copied second message to 0x0C24, then next image descriptor
;in your program is at 0x0E54 = 0x0C24 + 0x0230. This is done to make data "position
;independent", you can copy it to any address and don't lose access to data.
;-------------------------------------------------------------------------------
;#2: if in image descriptor pointer to image is < 0, then this image don't have
;a palette and when you'll negate this number you'll get this image's bpp (bits per
;pixel, may be 24 or 32), this is added for further compatiblity with that formats
;that support multiple images in one file and/or different bpp for each image (for
;example TIFF). And image data is located just after the structure.
;-------------------------------------------------------------------------------
include 'crtthumb.inc'

size_of_first_msg equ 24
IPC_pause	  equ 25
number_of_retries equ 16

IPC_name:
   call get_PID_and_load_path_for_IPC_usage
;now ebx contains the PID of receiver. Save it
   push ebx
   call load_image
   mov  eax,[aMeViewMode]
   cmp	ax,'TN'
   jne  .loaded
   shr	eax,16
   sub	al,'0'
   cmp	al,9
   jbe	@f
   sub	al,7
@@:
   xchg	al,ah
   sub	al,'0'
   cmp	al,9
   jbe	@f
   sub	al,7
@@:
   aad	16
   inc	eax
   push eax
   mov	edi,[_program_memory] ;put in the end of all memory
   mov	eax,256*256*4		  ;temporary solution, I don't like it
   push	eax
   mov	[thumbnail_mem],eax
   mov	[thumbnail],edi
   call mem_add_pages

   call CreateThumb
   mov  [thumbnail_size],eax  ;temporary solution, only for testing
   mov	eax,1
   jmp	.loaded_FillFirstTable
.loaded:
;build first msg. It'll be located at hash area (this temp area is huge enough)
;if some error occured we won't be here - program exited in that case in load_image proc.
   mov	eax,[___img_count]
   test eax,eax
   jnz	@f
   inc	eax
.loaded_FillFirstTable:
   mov	edi,hashtable
   stosd

   push eax		      ;for calculate_2nd_msg_size proc

   xor	eax,eax 	      ;flags
   stosd

   call calculate_2nd_msg_size
   stosd

   mov	eax,[IMG_INFO_BitsPerPoint]
   cmp  word[aMeViewMode],'TN'
   jne	.loaded_CopyParamsAsIs
   push eax
   xor	eax,eax
   stosd
   mov	eax,[thumbnail_size]
   stosd
   pop	eax
   mov	ax,32
   jmp	.loaded_CopyParams
.loaded_CopyParamsAsIs:
   mov	esi,IMG_INFO_Block
   movsd
   movsd
.loaded_CopyParams:   
   stosd

;send first msg
   pop	ecx		      ;receiver's PID
   mov	ebx,2		      ;send ipc message
   lea	edx,[edi-size_of_first_msg] ;it's shorter
   mov	esi,size_of_first_msg ;it's shorter
   push number_of_retries
.try_to_send_first_message:
   mov	eax,60
   int	0x40
;wait
   test eax,eax
   jz	.first_message_sent
   cmp	eax,3
   ja	exit_application      ;fatal errors. But it's not my fault
   dec	dword[esp]
   jz	exit_application      ;timeout. But it's not my fault
   push ebx
   mov	al,5
   mov	ebx,IPC_pause
   int	0x40
   pop	ebx
   jmp	.try_to_send_first_message

.first_message_sent:
   mov	[esp],ecx ;deallocate 'number_of_retries' var. and save receiver's PID

;build second msg. Remember that edx points to the first message
   mov	esi,[edx+8]	      ;Do you remember? This is sizeof(second_msg)
			      ;ebx==2, keep it in mind
   dec	dword[edx]
   jg	.second_message_with_multiple_images
   lea	eax,[edx+20]
   
   cmp  word[aMeViewMode],'TN'
   jne	@f
   mov	edx,[thumbnail]
   jmp	.try_to_send_second_message
@@:
   mov	edx,[soi]
   cmp	byte[eax],24
   jae	.try_to_send_second_message
   mov	eax,edx
   mov	edx,[___Palette]      ;due to the wise (not greedy!) memory using in
   sub	eax,edx 	      ;MeView palette is located just before the image data
   sub	edx,4
   mov	[edx],eax
   jmp	.try_to_send_second_message

;send second msg
.second_message_with_multiple_images:
   mov	edx,[___cur_info]     ;offset of resulting structure with images

.try_to_send_second_message:
   pop	ecx
   push number_of_retries
.try_to_send_second_message_retry:
   mov	eax,60
   int	0x40
;wait
   test eax,eax
   jz	exit_application      ;we've done
   cmp	eax,3
   ja	exit_application      ;fatal errors. But it's not my fault
   dec	dword[esp]
   jz	exit_application      ;timeout. But it's not my fault
   push ebx
   mov	al,5
   mov	ebx,IPC_pause
   int	0x40
   pop	ebx
   jmp	.try_to_send_second_message_retry

;-------------------------------------------------------------------------------
get_PID_and_load_path_for_IPC_usage:
   or	byte[status],bootmode ;it's common for all IPC ways of usage
   lea	esi,[eax+4]	      ;receiver's PID offset
   push 8
   pop	ecx
   xor	ebx,ebx
.1:
   lodsb
   sub	al,'0'
   cmp	al,9
   jbe	@f
   sub	al,7
@@:
   shl	ebx,4
   add	bl,al
   loop .1
;now ebx contains the PID of receiver. Save it
;esi points to the beginning of the path
   mov	edi,string	      ;copy string to its place
   mov	ecx,(256-12)/4
   rep	movsd
   ret


calculate_2nd_msg_size:
   cmp  word[aMeViewMode],'TN'
   jne	@f
   mov	eax,[thumbnail_mem]
   jmp	.loc_ret
@@:
   cmp	dword[esp+4],1
   ja	.more_than_one_image
   mov	eax,[IMG_INFO_Sizes]
   movzx ebx,ax
   shr	eax,16
   mul	ebx
.with_65th:		      ;we can have 1,2,4,8,15,16,24 or 32 bpp image
   movzx ebx,word[IMG_INFO_BitsPerPoint]
   cmp	ebx,15
   jne  @f
   inc	ebx                   ;because 15 bpp images have 2 bytes per pixel
@@:
   mul	ebx
   cmp	ebx,15
   jae	.loc_ret

   add	eax,[soi] ;if we have a palette, then we need space for it
   sub	eax,[___Palette]
   add	eax,4
.loc_ret:
   ret	4

.more_than_one_image:
   mov	eax,[output_data_size]
   jmp	.loc_ret
