;-------------------------------------------------------------------------------
;MeView thumbnails creation routines.
;Started: 21.11.2008
;Last changed: 26.01.2009
;-------------------------------------------------------------------------------
;Parameters:
;arg0 - required size (stdcall)
;edi  - offset of output area
;-------------------------------------------------------------------------------
;Returns:
;[edi] - 32bpp thumbnail, sizes are in eax, eax=NewWidth+NewHeight*65536
;-------------------------------------------------------------------------------
;During the resizing:
;esi - points to needed pixel from the original image
;edi - 32bpp output pointer
;ebp - palette pointer
;eax - current pixel
;ebx - NewWidth (for optimization)
;ecx - current x (in source image)
;edx - current y (in source image)
;-------------------------------------------------------------------------------
;Code
CreateThumb:
;local variables:
.NewHeightCounter	equ esp
.NewWidth			equ esp+4
.NewHeight			equ esp+8
.FractionX			equ esp+12
.FractionY			equ esp+16
.FractionStep		equ esp+20
.IntegerStep		equ esp+24
;.ReturnAddr		equ esp+28  ;just to keep in mind
.ReqSize			equ esp+28+4
	sub esp,28				;allocate locals
	
;warning! temporary code!
	mov ebp,[___Palette]
	xor edx,edx				;for dividing and some tricks
	mov ecx,[.ReqSize]		;ecx=ReqSize, it's common for both variants
	mov	dword[.FractionX],edx	;FractionX=0
	mov	dword[.FractionY],edx	;FractionY=0
	
	mov	eax,[IMG_INFO_Sizes];eax=width+height*65536
	movzx	ebx,ax			;ebx=width
	shr eax,16				;eax=height
	
	cmp ebx,eax				;if width>height
	ja	.WidthIsGreater		;go to variant2
;variant1 - height is >= width
	mov [.NewHeight],ecx	;NewHeight = ReqSize
	;NewWidth=Width*Ratio, where Ratio=ReqSize/Height
	xchg eax,ebx			;it's only 1 byte!
	mul	ecx				
	div	ebx					;eax=Width*ReqSize/Height
	mov [.NewWidth],eax		;NewWidth=eax
	mov	eax,ebx
	jmp .NewSizesCalculated
.WidthIsGreater:			;variant2
	mov [.NewWidth],ecx		;NewWidth = ReqSize
	;NewHeight=Height*Ratio, where Ratio=ReqSize/Width
	push eax
	mul	ecx				
	div	ebx					;eax=Width*ReqSize/Height
	mov [4+.NewHeight],eax	;NewHeight=eax (4+ because of saved eax)
	mov	eax,ebx
	pop	eax
.NewSizesCalculated:
	mov ebx,[.NewHeight]
	mov [.NewHeightCounter],ebx ;setup counter for outer loop (see below)
;Now we're going to calculate step size (dX and dY).
;Result will be in 32.32 fixed point format
	xor	edx,edx
	div	dword[.NewHeight]	;eaxis still == height
	mov [.IntegerStep],eax	;save high bits of quotient
	xor	eax,eax				;load low bits of dividend
	div dword[.NewHeight]
	mov [.FractionStep],eax	;now, qword at [.FractionStep] is step*(2^32)
;Now we are ready to create thumbnail. The only thing left is to write 2
;loops (second in the first) that will get pixels from original image
	xor	edx,edx
.InternalCycleForRows:
	mov	ebx,[.NewWidth]
	xor	ecx,ecx
	mov [.FractionX],ecx
.InternalCycleForColumns:
	call Prepare2Get
	call [GetPixel]
	stosd
;	test dword[.FractionX],-1
;	jns	@f
;	dec ecx					;rounding handling
;@@:	

	mov	eax,[.FractionStep]
	add [.FractionX],eax
	adc ecx,[.IntegerStep]

;	test eax,eax
;	jns	@f
;	inc ecx					;rounding handling
;@@:
	dec ebx
	jnz .InternalCycleForColumns

;	test dword[.FractionY],-1
;	jns	@f
;	dec edx					;rounding handling
;@@:	

	mov	eax,[.FractionStep]
	add [.FractionY],eax
	adc edx,[.IntegerStep]

;	test eax,eax
;	jns	@f
;	inc edx					;rounding handling
;@@:
	dec dword[.NewHeightCounter]
	jnz .InternalCycleForRows
;That's all. Calculate return value and return
	mov eax,[.NewHeight]
	shl eax,16
	mov ax,[.NewWidth]
	add esp,28
;warning! temporary code!
	ret 4