;-------------------------------------------------------------------------------
;MeView memory-managing functions.
;Created:      July 30, 2008
;Last changed: October 28, 2008
;-------------------------------------------------------------------------------
mem_add_pages:
;in: [arg0] - how much memory to add.
;out: nothing. If OUT_OF_MEMORY error, then error exit routine will be used
;(not implemented _YET_), control wouldn't be passed back to the routine
;that caused an error.
;Note, [arg0] means [esp+4], see stdcall convention, i.e. your call:
; push 4*4096 ;<---for example, add 4 pages (16384 bytes)
; call mem_add_pages
;
    push eax
    mov  eax,[_program_memory] ;current amount
    add  [esp+4+4],eax
    pop  eax
;continue...
mem_resize_everything:
;in: [arg0] - new memory size
;out: nothing. If OUT_OF_MEMORY error, then error exit routine will be used
;(not implemented _YET_), control wouldn't be passed back to the routine
;that caused an error.
;Note, [arg0] means [esp+4], see stdcall convention, i.e. your call:
; push new_mem_size
; call mem_resize_everything
;
    pushad
    mov eax,64
    mov ebx,1
    mov ecx,[esp+32+4]
    add ecx,4095      ;align to page boundary
    and ecx,not 0xFFF
    mov [_program_memory],ecx
    int 0x40
;    dec eax                ;if eax==1 then OUT_OF_MEMORY
;    jz  error_out_of_memory
    popad
    ret 4

;-------------------------------------------------------------------------------
align 4 ;this routine is called very often. Try to put it to cache.
mem_alloc_page_if_needed:
;in, out: nothing
;This routine is used to avoid page faults and other fatal exceptions
;when we can't calculate the final output size. It's used before 'stos' and
;'movs' instructions, if edi >= (memory_size-32) new page will be allocated.
;If we can't allocate new page (OUT_OF_MEMORY error) then error exit routine
;will be used (not implemented _YET_), control wouldn't be passed back to the
;routine that caused an error.
;
;mem_alloc_page_force just adds 1 page to our memory without any checking
;use it, for example, before building a palette.
    push edi
    sub  edi,[_program_memory]
    cmp  edi,-32
    pop  edi	      ;'pop' doesn't affect flags
    jl	 @f
mem_alloc_page_force:
    push 0x1000       ;mem_resize cares about aligning
    call mem_add_pages
@@:
    ret

;-------------------------------------------------------------------------------
;This routine deletes the part of compressed image data that has already
;been read and now is no longer needed
align 4 ;this routine is called very often. Try to put it to cache.
save_ram:
    push esi
    push edi
    push ecx
    mov  ecx,edi
    sub  ecx,esi
    mov  edi,[start_of_img_data]
    push ecx
    mov  ecx,esi
    sub  ecx,edi
    sub  [esp+8],ecx
    sub  [esp+12],ecx
    pop  ecx
    rep  movsb
    pop  ecx
    pop  edi
    pop  esi
    ret
